"""Custom logging class and setup function"""

import logging
from typing import Optional


def get_logger(name: str, level: Optional[int] = None) -> logging.Logger:
    """get or create a logger.

    Args:
        name (str)
        level (Optional[int]): Defaults to logging.DEBUG.

    Returns:
        logging.Logger
    """
    logger = logging.getLogger(name)

    logger.setLevel(logging.INFO if level is None else level)
    logger.propagate = False

    log_handler = logging.StreamHandler()
    log_handler.setFormatter(
        logging.Formatter("%(asctime)s %(levelname)s %(message)s", "%Y-%m-%d %H:%M:%S")
    )
    logger.addHandler(log_handler)

    return logger



import logging
import os
import socket
import time
from datetime import datetime
from typing import Optional

import requests
from croniter import croniter
from icloudpy import ICloudPyService
from icloudpy.exceptions import ICloudPyAPIResponseException
from icloudpy.services.photos import PhotoAsset
from requests.exceptions import ConnectionError  # pylint: disable=redefined-builtin

from icpd import constants


def wait_next(scheduler: croniter, logger: Optional[logging.Logger] = None) -> None:
    next_time: float = scheduler.get_next()
    sleep_sec: float = next_time - time.time()
    while sleep_sec <= 0:
        next_time = scheduler.get_next()
        sleep_sec = next_time - time.time()

    if logger:
        logger.info(
            "sleeping until %d (%d seconds)", round(next_time), round(sleep_sec)
        )

    remaining_sec: float = next_time - time.time()
    while remaining_sec > 0:
        time.sleep(3)
        remaining_sec = next_time - time.time()


def send_webhook(url: str, method: Optional[str], skip_verify: bool, data: str) -> None:
    verify: bool = not skip_verify
    if method and method.lower() == "get":
        resp = requests.get(url, verify=verify, timeout=10)
    else:
        resp = requests.post(
            url,
            verify=verify,
            json={
                "priority": 1,
                "title": "icpd cycle complete",
                "message": data,
            },
            timeout=10,
        )

    if not (200 <= resp.status_code < 400):
        raise Exception("error sending webhook; http error " + str(resp.status_code))


def download_media(
    icloud: ICloudPyService,
    logger: logging.Logger,
    photo: PhotoAsset,
    version: str,
    target_file: str,
) -> bool:
    """Download the photo to path, with retries and error handling"""

    download_dir: str = os.path.dirname(target_file)
    if not os.path.exists(download_dir):
        os.makedirs(download_dir, exist_ok=True)

    for retries in range(constants.MAX_RETRIES):
        try:
            photo_response = photo.download(version)
            if photo_response:
                temp_download_path = target_file + ".part"
                with open(temp_download_path, "wb") as file_obj:
                    for chunk in photo_response.iter_content(chunk_size=1024):
                        if chunk:
                            file_obj.write(chunk)
                temp_file_size = os.path.getsize(temp_download_path)
                if photo.versions[version]["size"] != temp_file_size:
                    raise Exception(
                        f"source file size ({photo.size}) != destination file size"
                        f" ({temp_file_size})"
                    )
                os.rename(temp_download_path, target_file)
                return True

            logger.error(
                "Could not find URL to download %s for size %s!",
                photo.filename,
                version,
            )
            break

        except (ConnectionError, socket.timeout, ICloudPyAPIResponseException) as ex:
            if "Invalid global session" in str(ex):
                logger.debug("Session error, re-authenticating...")
                if retries > 0:
                    # If the first reauthentication attempt failed,
                    # start waiting a few seconds before retrying in case
                    # there are some issues with the Apple servers
                    time.sleep(constants.WAIT_SECONDS)

                icloud.authenticate()
            else:
                # you end up here when p.e. throttling by Apple happens
                wait_time = (retries + 1) * constants.WAIT_SECONDS
                logger.error(
                    "Error downloading %s, retrying after %d seconds...",
                    photo.filename,
                    wait_time,
                )
                time.sleep(wait_time)
    else:
        raise Exception(
            "Could not download %s! Please try again later." % photo.filename
        )

    return False


def build_target_path(
    username: str,
    file_name: str,
    date_created: datetime,
    parent_directory: str,
    folder_structure: Optional[str] = None,
) -> str:
    date_path: str = ""
    if folder_structure:
        date_path = (
            folder_structure.format(date_created)
            if folder_structure.startswith("{") and folder_structure.endswith("}")
            else folder_structure
        )

    username_safe = "".join(x for x in username if x.isalnum())

    download_dir = os.path.normpath(
        os.path.join(parent_directory, username_safe, date_path)
    )

    return os.path.join(download_dir, file_name)



"""Get/set EXIF dates from photos"""
import struct

import piexif


def get_photo_exif(fpath: str):
    """Get EXIF date for a photo, return nothing if there is an error"""
    try:
        return piexif.load(fpath).get("Exif", {}).get(36867)
    except struct.error:
        pass
    return None


def set_photo_exif(fpath: str, date_str: str):
    """Set EXIF date on a photo"""
    exif_dict = piexif.load(fpath)
    exif_dict.get("1st")[306] = date_str
    exif_dict.get("Exif")[36867] = date_str
    exif_dict.get("Exif")[36868] = date_str
    exif_bytes = piexif.dump(exif_dict)
    piexif.insert(exif_bytes, fpath)



import os
from typing import Any, Dict

from icloudpy.services.photos import PhotoAsset

from icpd import constants


class PhotoAssetExt(PhotoAsset):
    # # https://github.com/picklepete/pyicloud/pull/347
    # PHOTO_VERSION_LOOKUP = {
    #     "full": "resJPEGFull",
    #     "large": "resJPEGLarge",
    #     "medium": "resJPEGMed",
    #     "thumb": "resJPEGThumb",
    #     "sidecar": "resSidecar",
    #     "original": "resOriginal",
    #     "original_alt": "resOriginalAlt",
    # }

    # VIDEO_VERSION_LOOKUP = {
    #     "full": "resVidFull",
    #     "medium": "resVidMed",
    #     "thumb": "resVidSmall",
    #     "original": "resOriginal",
    #     "original_compl": "resOriginalVidCompl",
    # }
    FILETYPE_EXTENSIONS = {
        "com.apple.quicktime-movie": "MOV",
        "public.heic": "HEIC",
        "public.jpeg": "JPG",
        "public.png": "PNG",
        "public.mpeg-4": "MP4",
    }

    def __init__(self, photo: PhotoAsset):
        super().__init__(photo._service, photo._master_record, photo._asset_record)

        asset_basename, asset_fileext = os.path.splitext(self.filename)

        # asset_record contains download links to edited versions;
        # master_record contains links to originals.
        self._versions: Dict[str, Any] = {}
        prefix: str = "res"
        suffix: str = "Res"
        # for record in (self._asset_record, self._master_record):
        for record in (self._master_record, self._asset_record):
            record_type: str = record["recordType"]
            fields: Dict[str, Any] = record["fields"]
            for k, v in fields.items():
                if k.startswith(prefix) and k.endswith(suffix):
                    v_name: str = k[: -len(suffix)]
                    v_meta: Dict[str, Any] = v["value"]

                    v_file_type = fields.get(v_name + "FileType", {})["value"]
                    v_file_ext = self.FILETYPE_EXTENSIONS.get(
                        v_file_type, asset_fileext
                    )

                    this_version: Dict[str, Any] = {
                        "filename": asset_basename
                        + (
                            ""
                            if record_type == constants.MASTER_RECORD_TYPE
                            else "-edit"
                        )
                        + "."
                        + v_file_ext.lstrip("."),
                        "size": int(v_meta["size"]),
                        "url": v_meta["downloadURL"],
                        "width": int(fields.get(v_name + "Width", {})["value"]),
                        "height": int(fields.get(v_name + "Height", {})["value"]),
                        "type": v_file_type,
                        "source": record_type,
                    }

                    if k in self._versions and self._versions[k] != this_version:
                        self._versions[f"{k}-edit"] = this_version
                    else:
                        self._versions[k] = this_version

        self._is_favorite: bool = bool(
            int(self._asset_record["fields"].get("isFavorite", {}).get("value", 0))
        )
        self._is_hidden: bool = bool(
            int(self._asset_record["fields"].get("isHidden", {}).get("value", 0))
        )

        self._is_edited: bool = (
            constants.EDITED_PHOTO in self._versions
            or constants.EDITED_VIDEO in self._versions
        )

        # these are mutually exclusive:
        self._is_live: bool = (
            constants.ORIGINAL_MEDIA in self._versions
            and constants.ORIGINAL_LIVE in self._versions
        )
        self._is_video: bool = (
            not self._is_live and constants.MEDIUM_VIDEO in self._versions
        )  # stills don't have MEDIUM_VIDEO
        self._is_still: bool = not self._is_live and not self._is_video

        self._media_type: str = (
            "live"
            if self._is_live
            else "video"
            if self._is_video
            else "still"
            if self._is_still
            else "unknown"
        )

        self._largest_versions: Dict[str, Any] = {}
        _largest_versions_lkp: Dict[str, int] = {}
        for k, v in self._versions.items():
            v_file_name: str = v["filename"]
            v_file_size: int = v["size"]
            if v_file_size > _largest_versions_lkp.get(v_file_name, 0):
                _largest_versions_lkp[v_file_name] = v_file_size
                self._largest_versions[v_file_name] = {k: v}
        self._largest_versions = {
            k2: v2 for k, v in self._largest_versions.items() for k2, v2 in v.items()
        }

    @property
    def versions(self) -> dict:
        return self._versions

    @property
    def largest_versions(self) -> dict:
        return self._largest_versions

    @property
    def is_favorite(self) -> bool:
        return self._is_favorite

    @property
    def is_hidden(self) -> bool:
        return self._is_hidden

    @property
    def is_edited(self) -> bool:
        return self._is_edited

    @property
    def is_live(self) -> bool:
        return self._is_live

    @property
    def is_still(self) -> bool:
        return self._is_still

    @property
    def is_video(self) -> bool:
        return self._is_video

    @property
    def media_type(self) -> str:
        return self._media_type



# pylint: disable=protected-access

import logging
import os
import re
import subprocess
import sys
from datetime import datetime, timedelta
from fnmatch import fnmatch
from re import Pattern
from typing import Any, Dict, List, Optional, Set, Tuple

import icloudpy
import icloudpy.services.photos
import prometheus_client as prom
import yapx
from croniter import croniter
from tabulate import tabulate

from . import constants, exif_datetime
from .__version__ import __version__
from .exceptions import ICloudLoginRequiredError, PhotoDownloadError
from .logger import get_logger
from .photo_asset_ext import PhotoAssetExt
from .utils import build_target_path, download_media, send_webhook, wait_next

try:
    from typing_extensions import Literal
except ModuleNotFoundError:
    from typing import Literal

GLOBAL_ARGS: Dict[str, Any] = {}

STATS_REGISTRY = prom.CollectorRegistry()

ALBUM_PHOTO_COUNT = prom.Gauge(
    name="license_activation_sum",
    documentation="license activation sum",
    labelnames=["username", "album", "stage"],
    registry=STATS_REGISTRY,
)
MEDIA_DOWNLOADED = prom.Counter(
    name="media_downloaded",
    documentation="count of downloaded media assets",
    labelnames=["username", "album", "type"],
    registry=STATS_REGISTRY,
)
MEDIA_DOWNLOADED_BYTES = prom.Counter(
    name="media_downloaded_bytes",
    documentation="size of downloaded media assets",
    labelnames=["username", "album", "type"],
    registry=STATS_REGISTRY,
)
MEDIA_DELETED = prom.Counter(
    name="media_deleted",
    documentation="count of deleted media assets",
    labelnames=["username", "album", "type"],
    registry=STATS_REGISTRY,
)


def setup(
    username: str = yapx.arg(env="ICPD_USERNAME"),
    session_directory: str = yapx.arg("~/.icloudpy", env="ICPD_SESSION_DIR"),
):
    GLOBAL_ARGS["username"] = [
        u.strip() if "@" in u else f"{u.strip()}@{constants.DEFAULT_EMAIL_DOMAIN}"
        for u in username.split(",")
    ]
    GLOBAL_ARGS["session_directory"] = os.path.expanduser(session_directory)


def icloud_login(_extra_args: Optional[List[str]] = None) -> bool:
    return all(
        subprocess.run(
            [
                "icloud",
                "--username",
                u,
                "--session-directory",
                GLOBAL_ARGS["session_directory"],
                *(_extra_args if _extra_args else []),
            ],
            check=False,
        ).returncode
        == 0
        for u in GLOBAL_ARGS["username"]
    )


def _get_icloud_session(username: str):
    icloud: Optional[icloudpy.ICloudPyService] = icloudpy.ICloudPyService(
        apple_id=username,
        cookie_directory=GLOBAL_ARGS["session_directory"],
    )

    if (
        icloud.requires_2fa or icloud.requires_2sa or not icloud.is_trusted_session
    ) and not icloud_login(_extra_args=["--non-interactive"]):
        raise ICloudLoginRequiredError()

    return icloud


def icloud_download(
    album: str = yapx.arg("All Photos", env="ICPD_ALBUM"),
    directory: str = yapx.arg(
        os.getcwd(), env="ICPD_OUTPUT_DIR", flags=["-d", "--directory"]
    ),
    download_count: int = yapx.arg(
        -1, env="ICPD_DOWNLOAD_COUNT", flags=["-n", "--download-count"]
    ),
    keep_count: int = yapx.arg(-1, env="ICPD_KEEP_COUNT", flags=["--keep-count"]),
    keep_days: int = yapx.arg(-1, env="ICPD_KEEP_DAYS", flags=["--keep-days"]),
    convert_heic: bool = yapx.arg(False, env="ICPD_CONVERT_HEIC"),
    delete_after_download: bool = yapx.arg(
        False, env="ICPD_DELETE", flags=["-x", "--delete-after-download"]
    ),
    oldest_first: bool = yapx.arg(False, env="ICPD_OLDEST_FIRST"),
    name_filter: Optional[str] = yapx.arg(None, env="ICPD_FILTER", flags=["--name"]),
    skip_album: Optional[List[str]] = yapx.arg(None, env="ICPD_SKIP_ALBUM"),
    skip_favorites: bool = yapx.arg(False, env="ICPD_SKIP_FAVORITES"),
    skip_hidden: bool = yapx.arg(False, env="ICPD_SKIP_HIDDEN"),
    skip_lives: bool = yapx.arg(False, env="ICPD_SKIP_LIVES"),
    skip_videos: bool = yapx.arg(False, env="ICPD_SKIP_VIDEOS"),
    skip_stills: bool = yapx.arg(False, env="ICPD_SKIP_STILLS"),
    break_if_exists: bool = yapx.arg(False, env="ICPD_BREAK_IF_EXISTS"),
    folder_structure: str = yapx.arg("{:%Y/%m}", env="ICPD_FOLDER_STRUCTURE"),
    webhook_url: Optional[str] = yapx.arg(None, env="ICPD_WEBHOOK_URL"),
    webhook_method: Optional[str] = yapx.arg(None, env="ICPD_WEBHOOK_METHOD"),
    webhook_skip_verify: bool = yapx.arg(False, env="ICPD_WEBHOOK_METHOD"),
    pre_sh: Optional[str] = yapx.arg(None, env="ICPD_PRE_SH"),
    post_sh: Optional[str] = yapx.arg(None, env="ICPD_POST_SH"),
    dry_run: bool = False,
    schedule: Optional[str] = yapx.arg(None, env="ICPD_SCHEDULE"),
    metrics_port: Optional[int] = yapx.arg(None, env="ICPD_METRICS_PORT"),
    log_level: Literal[
        "info", "debug", "warning", "error", "critical", "fatal"
    ] = "info",
) -> None:
    logger = get_logger(__name__, getattr(logging, log_level.upper()))

    scheduler: Optional[croniter] = None
    if schedule:
        logger.debug("cron-schedule: %s", schedule)
        scheduler = croniter(schedule)

    directory = os.path.normpath(os.path.expanduser(directory))

    if metrics_port and metrics_port > 0:
        logger.info("exposing metrics on port %d", metrics_port)
        prom.start_http_server(port=metrics_port, registry=STATS_REGISTRY)

    exit_code: int = 0
    while exit_code == 0:
        if pre_sh:
            logger.info("invoking: %s", pre_sh)
            subprocess.run(pre_sh, shell=True, check=True)

        for u in GLOBAL_ARGS["username"]:
            icloud: icloudpy.ICloudPyService = _get_icloud_session(username=u)

            if keep_count >= 0 or keep_days >= 0:
                oldest_first = True
                delete_after_download = True

            # instantiate this now so we can get any album and retrieve photos oldest-first.
            photos: icloudpy.services.photos.PhotoAlbum = icloud.photos.albums[album]

            expose_metrics: bool = (
                not dry_run and metrics_port is not None and metrics_port > 0
            )

            if expose_metrics:
                ALBUM_PHOTO_COUNT.labels(username=u, album=album, stage="before").set(
                    len(photos)
                )

            if oldest_first:
                logger.info("Configured to process oldest photos first.")
                for folder_settings in icloud._photos._albums.values():
                    folder_settings.direction = "DESCENDING"

            if delete_after_download is True:
                logger.info(
                    "Configured to delete source photos from iCloud after downloading!"
                )

            logger.debug("Looking up all photos and videos from album: %s", album)

            skip_photo_ids: Set[str] = set()

            if skip_album:
                for a in skip_album:
                    logger.debug("Obtaining file IDs from album: '%s'", a)
                    this_list = [p.id for p in icloud.photos.albums[a]]
                    skip_photo_ids.update(this_list)
                    logger.debug("Found %d files", len(this_list))

            photo_total_count: int = len(photos)
            photo_download_limit: int = photo_total_count
            if keep_count >= 0:
                photo_download_limit = max(0, photo_download_limit - keep_count)
            if download_count >= 0:
                photo_download_limit = min(download_count, photo_download_limit)

            logger.info(
                "Preparing to download %d of %d photos.",
                photo_download_limit,
                photo_total_count,
            )

            break_at_date: Optional[datetime] = None
            if keep_days > 0:
                break_at_date = datetime.combine(
                    datetime.utcnow().date() - timedelta(days=keep_days),
                    datetime.min.time(),
                )
                logger.info("Will ignore media created >= %s", str(break_at_date))

            logger.info("Destination directory: %s", directory)

            pattern: Optional[Pattern] = (
                re.compile(pattern=name_filter, flags=re.IGNORECASE)
                if name_filter
                else None
            )

            c: int = 0
            for i, item in enumerate(photos):

                logger.info(
                    "*** iteration=%d; downloads=%d; limit=%d",
                    i + 1,
                    c,
                    photo_download_limit,
                )
                if c >= photo_download_limit:
                    logger.info("Download limit reached: %d", photo_download_limit)
                    break

                remote_file: PhotoAssetExt = PhotoAssetExt(item)

                if break_at_date and remote_file.created >= break_at_date:
                    logger.info(
                        "Download limit reached: observed=%s; limit=%s",
                        remote_file.created,
                        break_at_date,
                    )
                    break

                if dry_run:
                    logger.info("*** (DRY RUN)")

                logger.info("remote file name: %s", remote_file.filename)

                # only look for original size of photos and videos.
                remote_file_versions: Dict[str, Any] = {
                    name: v
                    for name, v in remote_file.largest_versions.items()
                    if name in constants.DOWNLOAD_FORMATS
                }

                # only download "medium video" version if it is not from the original record
                # (this will download rendered versions of special-effect videos, e.g., slo-mo,
                # but not unecessarily download "medium video" in other cases.)
                if (
                    constants.MEDIUM_VIDEO in remote_file_versions
                    and remote_file_versions[constants.MEDIUM_VIDEO]["source"]
                    == constants.MASTER_RECORD_TYPE
                ):
                    del remote_file_versions[constants.MEDIUM_VIDEO]

                skip: bool = False
                if not remote_file_versions:
                    skip = True
                    logger.warning("No URL found for file.")
                elif skip_favorites and remote_file.is_favorite:
                    skip = True
                    logger.info("Skipping favorite media.")
                elif skip_hidden and remote_file.is_hidden:
                    skip = True
                    logger.info("Skipping hidden media.")
                elif (
                    (skip_lives and remote_file.is_live)
                    or (skip_videos and remote_file.is_video)
                    or (skip_stills and remote_file.is_still)
                ):
                    skip = True
                    logger.info("Skipping media type: %s", remote_file.media_type)
                elif skip_photo_ids and remote_file.id in skip_photo_ids:
                    skip = True
                    logger.info("Skipping media from an ignored album: %s", skip_album)
                elif pattern and not pattern.search(remote_file.filename):
                    skip = True
                    logger.info("Skipping media not matching filter.")
                # elif not (remote_file.is_edited and remote_file.is_video):
                #     raise Exception("idk")
                #     skip = True

                if skip:
                    continue

                success: List[bool] = []
                target_exists: bool = False
                for version_name, version_details in remote_file_versions.items():

                    logger.info("version: %s", version_name)

                    target_file: str = build_target_path(
                        username=u,
                        file_name=version_details["filename"],
                        date_created=remote_file.created,
                        parent_directory=directory,
                        folder_structure=folder_structure,
                    )

                    target_exists = os.path.exists(target_file)
                    if target_exists:
                        file_size = os.path.getsize(target_file)
                        photo_size = version_details["size"]

                        if file_size != photo_size:
                            target_file = f"-{photo_size}.".join(
                                target_file.rsplit(".", 1)
                            )
                            target_exists = False

                    if not target_exists:
                        logger.info("Downloading: %s, ", target_file)

                        if not dry_run and not download_media(
                            icloud=icloud,
                            logger=logger,
                            photo=remote_file,
                            version=version_name,
                            target_file=target_file,
                        ):
                            raise PhotoDownloadError(target_file)

                        if expose_metrics:
                            MEDIA_DOWNLOADED_BYTES.labels(
                                username=u,
                                album=album,
                                type=remote_file.media_type,
                            ).inc(version_details["size"])

                    if not dry_run:
                        conversion_output_files: List[str] = (
                            convert_heic_to_jpg(target_file)
                            if convert_heic and target_file.upper().endswith(".HEIC")
                            else []
                        )
                        for f in conversion_output_files + [target_file]:
                            if f.upper().endswith(
                                (".JPG", ".JPEG")
                            ) and not exif_datetime.get_photo_exif(f):
                                # %Y:%m:%d looks wrong but it's the correct format
                                date_str = remote_file.created.strftime(
                                    "%Y:%m:%d %H:%M:%S"
                                )
                                exif_datetime.set_photo_exif(f, date_str=date_str)

                    if target_exists:
                        logger.warning("Target already exists: %s", target_file)
                        if break_if_exists:
                            break

                    success.append(True)

                if delete_after_download is True:
                    if len(success) != len(remote_file_versions) or not all(success):
                        logger.info("Not all media was downloaded; not deleting.")
                    elif not dry_run:
                        logger.info("Deleting source file from iCloud.")
                        remote_file.delete()

                        if expose_metrics:
                            MEDIA_DELETED.labels(
                                username=u,
                                album=album,
                                type=remote_file.media_type,
                            ).inc(1)

                if expose_metrics:
                    MEDIA_DOWNLOADED.labels(
                        username=u,
                        album=album,
                        type=remote_file.media_type,
                    ).inc(1)

                c += 1

                if target_exists and break_if_exists:
                    logger.info("Stopping early.")
                    break

            if expose_metrics:
                photos._len = None  # reset to not rely on cached count.
                ALBUM_PHOTO_COUNT.labels(username=u, album=album, stage="before").set(
                    len(photos)
                )

        if post_sh:
            logger.info("invoking: %s", post_sh)
            subprocess.run(post_sh, shell=True, check=True)

        logger.info("Done.")

        if webhook_url and not dry_run:
            send_webhook(
                url=webhook_url,
                method=webhook_method,
                skip_verify=webhook_skip_verify,
                data="complete",
            )

        if not scheduler:
            break
        if exit_code == 0:
            wait_next(scheduler, logger=logger)
    # END: while

    logger.debug("exit code: %d", str(exit_code))
    sys.exit(exit_code)


def convert_heic_to_jpg(path: str, quality: int = 100) -> List[str]:
    source_file_list: List[str] = (
        [
            os.path.join(dir_name, x)
            for dir_name, filter_to in [os.path.split(path)]
            for x in os.listdir(dir_name)
            if x.lower().endswith(".heic") and fnmatch(x, filter_to)
        ]
        if "*" in path
        else [path]
        if path.lower().endswith(".heic")
        else [
            os.path.join(path, x)
            for x in os.listdir(path)
            if x.lower().endswith(".heic")
        ]
        if os.path.isdir(path)
        else []
    )

    target_files: List[Tuple[str, str]] = [
        (x, os.path.splitext(x)[0] + ".jpg") for x in source_file_list
    ]

    for f, f_new in target_files:
        subprocess.run(
            ["heif-convert", "-q", str(quality), f, f_new],
            check=True,
        )
        p: subprocess.CompletedProcess = subprocess.run(
            ["exiftool", "-n", "-QuickTime:Rotation", f], check=False, text=True
        )

        exiftool_args: List[str] = [
            "exiftool",
            "-overwrite_original",
            "-P",
            "-n",
            "-ModifyDate<FileModifyDate",
        ]

        if p.returncode == 0:
            exiftool_args.append("-Orientation=1")
        subprocess.run(exiftool_args + [f_new], check=True)

    return [x for _, x in target_files]


def _list_albums(
    username: str,
    filter_to: Optional[str] = None,
    stats: bool = False,
    _icloud: Optional[icloudpy.ICloudPyService] = None,
) -> List[Dict[str, Any]]:
    if not _icloud:
        _icloud = _get_icloud_session(username=username)

    return [
        {"Album": str(v), "Photos": len(v)} if stats else {"Album": str(v)}
        for v in _icloud.photos.albums.values()
        if not filter_to or fnmatch(str(v), filter_to)
    ]


def list_albums(
    filter_to: Optional[str] = yapx.arg(None, flags=["--filter"]), stats: bool = False
) -> None:
    for u in GLOBAL_ARGS["username"]:
        print(
            tabulate(
                _list_albums(username=u, filter_to=filter_to, stats=stats),
                headers="keys",
                tablefmt="github",
            )
        )


def main() -> None:
    yapx.run(
        setup,
        **{
            "version": lambda: print(__version__),
            "login": icloud_login,
            "convert-heic": convert_heic_to_jpg,
            "albums": list_albums,
            "download": icloud_download,
        },
        _args=sys.argv[1:],
    )


if __name__ == "__main__":
    main()



"""Constants"""

from typing import List

DEFAULT_EMAIL_DOMAIN = "icloud.com"

# For retrying connection after timeouts and errors
MAX_RETRIES = 5
WAIT_SECONDS = 5

MASTER_RECORD_TYPE = "CPLMaster"

# original photos and videos have this version.
ORIGINAL_MEDIA = "resOriginalRes"

# only live photos have this version.
ORIGINAL_LIVE = "resOriginalVidComplRes"

# only videos have this version.
# videos with special effects (e.g., slo-mo)
MEDIUM_VIDEO = "resVidMedRes"

# edited photos and video thumbnails have this version.
EDITED_PHOTO = "resJPEGFullRes"
EDITED_VIDEO = "resVidFullRes"

DOWNLOAD_FORMATS: List[str] = [
    ORIGINAL_MEDIA,
    ORIGINAL_LIVE,
    EDITED_PHOTO,
    EDITED_VIDEO,
    MEDIUM_VIDEO,
]
DOWNLOAD_FORMATS.extend([x + "-edit" for x in DOWNLOAD_FORMATS])
